# -*- coding: utf-8 -*-

# This is part of the pEp for Thunderbird Setup
# Copyright (C) 2020, pEp Foundation, Switzerland
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 2
# of the License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin St., 5. Floor, Boston, 02110-1301, USA.


from __future__ import print_function, unicode_literals, absolute_import

import errno
import fcntl
import os
import socket
import sys


class MozillaProfileLock(object):

    if sys.platform == "darwin":
        LOCKFILE_NAME = ".parentlock"
        OLD_LOCKFILE_NAME = "parent.lock"
    elif sys.platform != "win32":
        LOCKFILE_NAME = ".parentlock"
        OLD_LOCKFILE_NAME = "lock"
    else:
        LOCKFILE_NAME = "parent.lock"
        OLD_LOCKFILE_NAME = None

    def __init__(self, profile):
        self.profile = profile
        self.haveLock = 0
        self.lockFileDesc = None
        self.pidLockFilename = None
        self.errno = 0

    def lockWithFcntl(self, lockfile):
        try:
            fd = os.open(lockfile,
                    os.O_WRONLY | os.O_CREAT | os.O_TRUNC | os.O_NOFOLLOW, 0o666)
        except (OSError, IOError) as e:
            self.errno = e.errno
            return False
        try:
            fcntl.lockf(fd, fcntl.LOCK_EX | fcntl.LOCK_NB, 0, 0)
        except IOError as e:
            os.close(fd)
            self.errno = 0 if e.errno in (errno.EACCES, errno.EAGAIN) else e.errno
            return False
        self.lockFileDesc = fd
        self.errno = 0
        self.haveLock = 1
        return True

    def getLocalAddr(self):
        a = '127.0.0.1'
        try:
            a = socket.gethostbyname_ex(socket.gethostname())[2][0]
        except socket.gaierror:
            pass
        return a

    def isLockSymlinkStale(self, lockfile, haveFcntl=False):
        try:
            host, pid = os.readlink(lockfile).split(':', 1)
        except OSError as e:
            if e.errno == errno.EINVAL:
                return False
        self.errno = 0
        try:
            socket.inet_aton(host)
        except OSError as e:
            self.errno = e.errno
            return False
        if host != self.getLocalAddr():
            return False
        if pid.startswith('+') and haveFcntl:
            return True
        pid = int(pid)
        if pid == os.getpid():
            return False
        try:
            os.kill(pid, 0)
        except OSError as e:
            if e.errno == errno.ESRCH:
                return True
            raise
        return False

    def lockWithSymlink(self,  lockfile, haveFcntl=False):
        a = self.getLocalAddr()
        signature = "{0}:{1}{2:d}".format(a, "+" if haveFcntl else "", os.getpid())
        symlink_not_stale = False
        tries = 100
        while tries > 0:
            symlink_not_stale = False
            try:
                os.symlink(signature, lockfile)
                break
            except OSError as e:
                self.errno = e.errno
                if e.errno != errno.EEXIST:
                    break
            if not self.isLockSymlinkStale(lockfile, haveFcntl):
                symlink_not_stale = True
                break
            try:
                os.unlink(lockfile)
            except OSError as e:
                self.errno = e.errno
            tries -= 1
        if symlink_not_stale or not tries:
            return False
        self.pidLockFilename = lockfile
        self.haveLock = 1
        return True

    def lock(self):
        if self.haveLock > 0:
            self.haveLock += 1
            return True
        profile_dir = self.profile
        lockfile = os.path.join(profile_dir, self.LOCKFILE_NAME)
        if os.path.isdir(lockfile):
            return False
        if sys.platform != "win32":
            oldlockf = os.path.join(profile_dir, self.OLD_LOCKFILE_NAME)
            fcntl_locked = self.lockWithFcntl(lockfile)
            if not fcntl_locked and not self.errno:
                return False
            return self.lockWithSymlink(oldlockf, not bool(self.errno))
        else:
            # Use Python 3.3+ 's 'x' mode if possible to avoid race condition
            if sys.version_info[0:3] >= (3, 3, 0):
                try:
                    self.lockFileDesc = os.open(lockfile, 'x')
                except FileExistsError:
                    self.lockFileDesc = None
            else:
                if not os.path.exists(lockfile):
                    self.lockFileDesc = os.open(lockfile, 'w')
            return self.lockFileDesc is not None

    def unlock(self):
        if self.haveLock == 1:
            fd = self.lockFileDesc
            if fd is not None:
                os.close(fd)
                self.lockFileDesc = None
            if self.pidLockFilename:
                os.unlink(self.pidLockFilename)
        self.haveLock -= 1
        return True
